// Unit.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import java.io.*;

import swarm.gui.Raster;
import swarm.gui.ZoomRasterImpl;

/**
 * The Unit class instances are micro unit in our virtual enterprise;
 * i.e. the units where the steps required to build a product (to fulfill
 * an order) are done
 *
 * @author Pietro Terna (with Marco Remondino, Dario Landini, Nicola 
 * Ormezzano and Paolo Pelligra contributions)
 */
public class Unit extends SwarmObjectImpl {

    /** the number identifying the unit*/
    public int unitNumber;
    /** the unit computationalAssembler */
    ComputationalAssembler myComputationalAssembler;
    /** the unit production phase */
    public int unitProductionPhase;

    /** the list of the orders to be executed */
    public ListImpl  waitingList;
    /** the list of the orders in the waitingList, not kept into
     *  a parallel unit, such as a myComputationalAssembler */
    public ListImpl usableWaitingList;
    /** the list containing the orders (if any) performed in a day */
    public ListImpl dailyProductionList;
    /** the modelSwarm (to which we apply the 
     * setProductionTimeAndRecipeLength(...) method */
    ESFrameModelSwarm eSFrameModelSwarm;
    /** the temporary address of an existing unit */
    public Unit aUnit;
    /** the first order to be executed */
    public Order firstOrder;
    /** position of the unit into the display */
    public int xPos, yPos; 
   /** position of the unit into the display */
    public byte unitColor;
    /** the unitActivitySpace */
    public UnitActivitySpace unitActivitySpace;
    /** our visibility */
    public double visibility;
    /** increasing visibility in each step 1 */
    public double increasingVisibilityStep;
    /** the number of tick of inactivity */
    public int inactivity;
    /** unsent product at the end of step2 */
    public int unsentProducts;
    /** if droppable is true the unit will be dropped, regardless all
     *  its values (this come from a computational step) */
    public boolean droppable;
    /** out memory matrix */
    public MemoryMatrix myMemoryMatrix;
    /** max number of ticks that an order can stay in the unit */
    public int maxTicksAsUnsent;
    /** shortcut in case of same unit LIFO assignement */
    public boolean shortcutForLIFOInternalAssignement;

    /** the instance of AssigningTool we use */
    AssigningTool assigningTool; 

    /**
     * the constructor for Unit
     */
    public Unit (Zone aZone, ESFrameModelSwarm mo, UnitActivitySpace uas, 
		 AssigningTool at, int x, int y, byte col, double vis, double visStep,
		 ComputationalAssembler ca, MemoryMatrix mm, int mt) {

	// Call the constructor for the Unit parent class
	super(aZone);

	init(aZone, mo, uas, 
		 at, x, y, col, vis, visStep,
		 ca, mm, mt);
    }

    /** managing the constructor content */
    public void init (Zone aZone, ESFrameModelSwarm mo, UnitActivitySpace uas, 
		 AssigningTool at, int x, int y, byte col, double vis, double visStep,
		 ComputationalAssembler ca, MemoryMatrix mm, int mt)
    {

	// the production waiting list of the unit
	waitingList = new ListImpl (getZone());
	// the the list of the orders in the waitingList, not kept into
	//  any parallel unit, such as (in jESOF) myComputatationalAssembler
	usableWaitingList = new ListImpl (getZone());
	// the model
	eSFrameModelSwarm = mo;
	// the daily list of finished orders
	dailyProductionList= new ListImpl(getZone());
	// the assigning tool
	assigningTool=at;
	// x and y pos
	xPos=x;
	yPos=y;
	// the color
	unitColor=col;
	// the unit activity space
	unitActivitySpace = uas;
	// our visibility
	visibility = vis;
	// increasing visibility
	increasingVisibilityStep=visStep;
	// our computational assembler
	myComputationalAssembler = ca;  
	// our memoruy matrix
	myMemoryMatrix = mm;
	// initial number of inativity ticks
	inactivity=0;
	// initial value of unsent products
	unsentProducts=0;
	// max ticks number that an order can spend in the unit
	maxTicksAsUnsent=mt;
	// setting false to droppable
	droppable=false;
	shortcutForLIFOInternalAssignement=false;

   }

    /** set unit number */
    public void setUnitNumber(int un)
    {
	// the unit id number.
	unitNumber=un;
	// announce the unit to the console
	if(StartESFrame.verbose)
	System.out.println("Unit number " + unitNumber + " has been created.");
     }

    /** get unit number */
    public int getUnitNumber()
    {
	// the unit id number.
	return unitNumber;
    }


    /** copying the orders in the waitingList, if they are not kept into
     *  any parallel unit, such as (in jESOF) myComputationalAssembler */
    public void setUsableWaitingList(){
	int i;
	Order tmp;

	// voiding usableWaitingList
	while(usableWaitingList.getCount()>0)usableWaitingList.removeFirst();
	if(waitingList.getCount()==0)return;
	for(i=0;i<waitingList.getCount();i++)
	    {
		tmp=(Order) waitingList.atOffset(i);
		if(! myComputationalAssembler.thisOrderIsInTheWaitingList(tmp))
		    usableWaitingList.addLast(tmp);
	    }

    }

    /**
     * production:<br><br>
     *
     * at the beginning of each time tick in the 'day', each unit looks for 
     * the order(s) (if any) to be executed; the production 
     * cycle is 'one unit of time'
     * in a step of the recipe of an order in a tick of a day;<br><br>
     */
    public void unitStep1 ()
    {
	int i, ii, cycle; 

	if(StartESFrame.verbose)
	    System.out.println("Step 1 in unit " + unitNumber +
			       "; waitingList count " + 
			       waitingList.getCount());

	// showing the visibility
	unitActivitySpace.showVisibility(xPos, yPos, (int) visibility);

	// increasing visibility
	visibility += increasingVisibilityStep;

	// inactivity
	if(waitingList.getCount() == 0) inactivity++; // usableWaitingList can be void,
	                                              // but a computational step can be kept here
	if(eSFrameModelSwarm.getMaxInactivity()==0)
	    unitActivitySpace.showDownHistogram(xPos, yPos, 0);
	else
	    unitActivitySpace.showDownHistogram(xPos, yPos, inactivity);

	cycle=0;
	//
	// repeating the cycle of unitStep1 if we find zero time orders
	//
	while(cycle==0){
	    cycle++; // cycle is reset to zero below

	    // preparing the sub lists
	    setUsableWaitingList();

	    // resetting cycle to zero to repeat it if a zero time order is in first position
	    if(usableWaitingList.getCount()>0)
		if(((Order)usableWaitingList.atOffset(0)).checkZeroTimeProduction())cycle=0;

	    // the production

	    // after the specific step in recipe production we place the order(s) 
	    // in a daily production
	    // list, which will be used at unitStep2 to diffuse production
	    // results

	    if(usableWaitingList.getCount() > 0){
		inactivity=0;
		firstOrder=(Order) usableWaitingList.removeFirst();
		waitingList.remove(firstOrder);
		
		if(firstOrder.getNextStep()!=getProductionPhase()){
		    System.out.println("Unrecoverable error (case2) in unit " 
				       + unitNumber + " (order number "
				       + firstOrder.getOrderNumber()+ " step "
				       + firstOrder.getNextStep()+")");

		    MyExit.exit(2);}
			
		firstOrder.setDoneStep(unitNumber);

		// redo zero time steps immediately
		if(InformationRepository.getRedoZeroTimeStepsImm() &&
		   firstOrder.checkZeroTimeProduction() && 
		   getProductionPhase()==firstOrder.getNextStep() &&
		   myComputationalAssembler. // the computational step does not exist
                                             // or can be immediately done
		   checkingComputationsOnASpecificOrder(firstOrder))
		    { 
			cycle=0;
			usableWaitingList.addFirst(firstOrder);    
			firstOrder.setNumberOfTicksAsUnsent(0);
			eSFrameModelSwarm.addToGlobalProduction(1);    
		    }
		else
		    {
			dailyProductionList.addLast(firstOrder);
			firstOrder.setNumberOfTicksAsUnsent(0);
			eSFrameModelSwarm.addToGlobalProduction(1);
		    }
	    }
	    
	} // concluding cycle
    } // concluding step1


    /**
     * diffusion and accounting:<br><br>
     *
     * sending the orders to the subsequent production units;<br><br>
     *
     * making accounting operations
     */
    public void unitStep2 () {
	float timeSpent;
	int numberOfProd, product;
	Order notAssignedOrder;

	// the leftHistogram related to the waitingList (with computational steps
	// not included in usableWaitingList)
	unitActivitySpace.showLeftHistogram(xPos, yPos, getWaitingListLength());
	    

	if(StartESFrame.verbose)
	    System.out.println("Step 2 in unit " + unitNumber +
			       " dailyProductionList count " +
			       dailyProductionList.getCount());

	// eliminating orders from the dailyProductionList if they are waiting
	// for more than maxTicksAsUnsent
	checkDailyProductionListForOldUnsentOrders();

	numberOfProd=dailyProductionList.getCount();
	for (product=0;product<numberOfProd;product++)
	    {
		firstOrder=(Order)dailyProductionList.removeFirst();

		// if the next step of the order as to be executed by the same unit, we
		// reassign it directly to this unit in a LIFO way
		if(InformationRepository.getSameUnitLifoAssignment() &&
		   getProductionPhase()==firstOrder.getNextStep()) // getNextStep()==0 
		                                                   // if finished
		    {
			shortcutForLIFOInternalAssignement=true;
			setWaitingList(firstOrder);
		    }
		else
		    {
			// do we have concluded the production step?
			// if yes ... we drop the object
			if (firstOrder.getNextStep()==0)
			    {
				if(StartESFrame.verbose)
				    System.out.println("drop order # " +
						       firstOrder.getOrderNumber() +
						       " startingTimeUnit " +
						       firstOrder.getStartingTimeUnit() +
						       " endTimeUnit " +
						       Globals.env.getCurrentTime() +
						       " stepNumber " +
						       firstOrder.getStepNumber() +
						       " ratio " +
						       ( (float)
							 (Globals.env.getCurrentTime() -
							  firstOrder.getStartingTimeUnit()))/
						       firstOrder.getStepNumber());
				// the production is concluded, so we inform
				// our environment (the model) of the time spent
				// and of the # of steps in recipe
				timeSpent=Globals.env.getCurrentTime() -
				    firstOrder.getStartingTimeUnit();
				eSFrameModelSwarm.setProductionTimeAndRecipeLength(
							     timeSpent,
							     firstOrder.getNumberOfDirectStepsDone());
				if(StartESFrame.verbose)
				    System.out.println("Order #" +
						       firstOrder.orderNumber +
						       " has been completed; time spent in ticks: " +
						       timeSpent +
						       " expected time in ticks " +
						       firstOrder.getStepNumber() + 
						       " or with stand alone batches " + 
						       firstOrder.getNumberOfDirectStepsDone());
				
				// adding the production to the concluded order value
				eSFrameModelSwarm.addToConcludedOrders(firstOrder.
								       getNumberOfDirectStepsDone());
				if(StartESFrame.verbose)
				    System.out.println("concluded " + 
						       firstOrder.getNumberOfDirectStepsDone());
				firstOrder.drop();
			    }
			
			// if not ... we send it to the subsequent production unit
			// if not found, the order is placed into the dailyProductionList
			else 
			    {
				notAssignedOrder=assigningTool.assign(this, firstOrder);
				if (notAssignedOrder != null)dailyProductionList.
								 addLast(notAssignedOrder);
			    }
		    }
	    }

	// the rightHistogram related to the dailyProductionList (here the value is
	// that un unsent products)
	unsentProducts = dailyProductionList.getCount();
	if(eSFrameModelSwarm.getMaxUnsentProducts()==0)
	    unitActivitySpace.showRightHistogram(xPos, yPos, 0);
	else
	    unitActivitySpace.showRightHistogram(xPos, yPos, unsentProducts);

    } // concluding step2

    /** set unit production phases */
    public void setProductionPhase(int ph)
    {
	unitProductionPhase = ph;
    }


    /**
     * return the production phases of the unit
     */
    public int getProductionPhase()
    {
	return unitProductionPhase;
    }

    /**
     * Putting an order in the production waitingList of the unit
     * (using a related object if any)
     */
    public void setWaitingList(Order anOrder)
    {
	int i; 

       // case "c": the order is included in myComputationalAssembler
       // waitingList; until it is in this list, it cannot be processed
       // in unitStep1 method
       // if commputations are immediately completed, the order will be
       // immediately freed in the successive schedule step
       if(anOrder.checkIfNextStepUsesAComputationalSpecificationSet())
	   {
	       myComputationalAssembler.setComputationalWaitingList(anOrder);
	       InformationRepository.setComputationalStepsInUse(true);
	   }


       // normal case: 
       //              orders with 0 time are assigned LIFO
       //
       //              orders coming from the same step execution (look at
       //              assigningTool) are assigned LIFO if 
       //              sameUnitLifoAssignment is true
       if(anOrder.checkZeroTimeProduction())waitingList.addFirst(anOrder);

       else if (shortcutForLIFOInternalAssignement)waitingList.addFirst(anOrder);
       else waitingList.addLast(anOrder);

	if(StartESFrame.verbose)
	    System.out.println("The unit " + unitNumber +
			       " has received the order # "
			       + anOrder.getOrderNumber());

	shortcutForLIFOInternalAssignement=false;
    }

    /** the visibility */
    public int getVisibility()
    {
	return (int) visibility;
    }

    /** the inactivity */
    public int getInactivity()
    {
	return inactivity;
    }

    /** the X position */
    public int getXPos()
    {
	return xPos;
    }

    /** the Y position */
    public int getYPos()
    {
	return yPos;
    }

    /** the length of the wainting list */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }
    


    /** unsent products at the end of step 2 */
    public int getUnsentProducts()
    {
	return unsentProducts;
    }

    /** get daily production list */
    public ListImpl getDailyProductionList(){
	return dailyProductionList;
    }

    /** get waiting list */
    public ListImpl getWaitingList(){
	return waitingList;
    }

    /** designing itself on the raster */
    public Object drawSelfOn (Raster r)
    {
	r.drawPointX$Y$Color (xPos, yPos, unitColor);
	return this;
    }

    /** getting the model the unit belongs to */
    public Object getUnitModel()
    {
	return eSFrameModelSwarm;
    }

    /** set the droppability of the unit */
    public void setDroppable()
    {
	droppable=true;
    }

    /** get the droppability of the unit */
    public boolean getDroppable()
    {
	return droppable;
    }

    /** get the memory matrix of the unit */
    public MemoryMatrix getMemoryMatrix()
    {
	return myMemoryMatrix;
    }

    /** get the computational assembler of the unit */
    public ComputationalAssembler getComputationalAssembler()
    {
	return myComputationalAssembler;
    }

    /** eliminating old order */
    public void checkDailyProductionListForOldUnsentOrders()
    {
	int i, k, listCount;
	Order tmp;

	if(maxTicksAsUnsent==0)return;
	listCount=dailyProductionList.getCount();
	if(listCount==0)return;
	
	for(i=0;i<listCount;i++)
	    {
		tmp=(Order) dailyProductionList.removeFirst();
		k=tmp.getNumberOfTicksAsUnsent();
		tmp.setNumberOfTicksAsUnsent(++k);

		if(maxTicksAsUnsent>k)		
		    dailyProductionList.addLast(tmp);
		else tmp.drop();
	    }
    }

    /** print the memory matrix of the unit (to be used via the probe
     *  of the unit) */
    public void printDirectlyMemoryMatrix()
    {
	myMemoryMatrix.printDirectly();
    }

    /** virtual dropping function, to re-use the unit */
    public void drop()
    {
	InformationRepository.addToHibernatedUnitList(this);
    }

}
